import pytest
from HLR.regression import HierarchicalLinearRegression
import pandas as pd

# Example test data
df = pd.DataFrame({
    'PTS': [120, 115, 130, 112, 104],
    'ORB': [10, 11, 8, 10, 9],
    'BLK': [5, 3, 6, 7, 4],
    'W': [50, 45, 55, 47, 44]
})
X = {1: ['PTS'], 2: ['PTS', 'ORB'], 3: ['PTS', 'ORB', 'BLK']}
y = 'W'

def test_model_initialization():
    """Test initialization of HierarchicalLinearRegression model."""
    hlr = HierarchicalLinearRegression(df, X, y)
    assert hlr.data is not None
    assert hlr.outcome_var == y

def test_fit_models():
    """Test fitting models."""
    hlr = HierarchicalLinearRegression(df, X, y)
    model_results = hlr.fit_models()
    assert len(model_results) == len(X)


"""HLR - Hierarchical Linear Regression."""

__author__ = """Toomas Erik Anijärv"""
__email__ = 'toomaserikanijarv@gmail.com'
__version__ = '0.2.0'

from .diagnostic_tests import (test_durbin_watson, 
                               test_pearsons_r,
                               test_rainbow, 
                               test_breusch_pagan,
                               test_goldfeld_quandt, 
                               test_pairwise_correlations,
                               test_vif, 
                               test_extreme_standardized_residuals,
                               test_cooks_distance, 
                               test_mean_of_residuals,
                               test_shapiro_wilk)
from .plots import (create_subplot_residuals_vs_fitted, 
                    create_subplot_qq_residuals,
                    create_subplot_influence, 
                    create_subplot_std_residuals,
                    create_subplot_histogram_std_residuals, 
                    create_subplot_partial_regression)
from .regression import HierarchicalLinearRegression



"""Functions for running hierarchical linear regression.

Authors
-------
Toomas Erik Anijärv toomaserikanijarv@gmail.com github.com/teanijarv
Rory Boyle rorytboyle@gmail.com github.com/rorytboyle
"""
import pandas as pd
import statsmodels.api as sm
import scipy.stats
import numpy as np
import pingouin as pg
import matplotlib.pyplot as plt

from . import diagnostic_tests
from . import plots

class HierarchicalLinearRegression:
    """Class for performing hierarchical linear regression analysis."""

    def __init__(self, df, ivs_dict, dv):
        """Initializes the HierarchicalLinearRegression class.

        Args:
            df (pd.DataFrame): The dataset to be used in the regression.
            ivs_dict (dict): Dictionary mapping model levels to lists of independent variables.
            dv (str): The dependent variable.
        """
        if not isinstance(df, pd.DataFrame):
            raise ValueError("Data is not a pandas DataFrame.")
        self.data = df

        if not isinstance(ivs_dict, dict):
            raise ValueError("ivs_dict must be a dictionary.")
        if not all(isinstance(k, int) and isinstance(v, list) for k, v in ivs_dict.items()):
            raise ValueError("ivs_dict keys must be integers and values must be lists.")
        if not all(col in df.columns for predictors in ivs_dict.values() for col in predictors):
            raise ValueError("Predictors in ivs_dict are not columns in DataFrame.")
        self.models = ivs_dict
        
        if dv not in df.columns:
            raise ValueError(f"{dv} is not a column in the DataFrame.")
        self.outcome_var = dv

    def fit_models(self):
        """Fits OLS models for each level of independent variables.

        Returns:
            dict: A dictionary of statsmodels OLS regression results.
        """
        results = {}
        for level, predictors in self.models.items():
            X = self.data[predictors]
            X_const = sm.add_constant(X)
            y = self.data[self.outcome_var]
            model = sm.OLS(y, X_const).fit()
            results[level] = model
        return results

    def calculate_change_stats(self):
        """Calculates change statistics between OLS models.

        Returns:
            dict: A dictionary containing change statistics.
        """
        stats = {}
        prev_model = None
        for level, curr_model in self.fit_models().items():
            if prev_model:
                pred_change = len(curr_model.model.exog[0]) - len(prev_model.model.exog[0])
                r_sq_change = curr_model.rsquared - prev_model.rsquared
                f_change = (r_sq_change / pred_change) / ((1 - curr_model.rsquared) / curr_model.df_resid)
                f_change_pval = scipy.stats.f.sf(f_change, pred_change, curr_model.df_resid)

                stats[level] = {
                    'R-squared change': r_sq_change,
                    'F-value change': f_change,
                    'P-value (F-value change)': f_change_pval
                }
            prev_model = curr_model
        return stats
    
    def calculate_additional_stats(self, model, X, y):
        """Calculates additional statistics for a given OLS model.

        Args:
            model (RegressionResults): The regression model.
            X (pd.DataFrame): Independent variables.
            y (pd.DataFrame): Dependent variable.

        Returns:
            dict: A dictionary containing additional statistics.
        """
        std_x = X.apply(np.std, ddof=1)
        std_y = y.apply(np.std, ddof=1).iloc[0]
        conv_factors = std_x / std_y
        beta_coeffs = pd.Series(model.params).drop('const')
        std_beta_coefs = (beta_coeffs * conv_factors).to_dict()

        Xy_temp = pd.concat([X, y], axis=1)
        partials = {col: pg.partial_corr(data=Xy_temp, x=col, y=y.columns[0], 
                                         covar=[x for x in X.columns if x != col])['r'].iloc[0] for col in X.columns}
        semi_partials = {col: pg.partial_corr(data=Xy_temp, x=col, y=y.columns[0], 
                                              x_covar=[x for x in X.columns if x != col])['r'].iloc[0] for col in X.columns}
        unique_var = {col: value ** 2 * 100 for col, value in semi_partials.items()}

        return {
            'Std Beta coefs': std_beta_coefs,
            'Partial correlations': partials,
            'Semi-partial correlations': semi_partials,
            'Unique variance %': unique_var
        }

    def summary(self):
        """Generates a summary report of the hierarchical linear regression analysis.

        Returns:
            pd.DataFrame: A DataFrame containing the summary report.
        """
        model_res = self.fit_models()
        change_stats = self.calculate_change_stats()

        report = []
        for level, model in model_res.items():
            X = self.data[model.model.exog_names[1:]]
            y = self.data[[self.outcome_var]]

            add_stats = self.calculate_additional_stats(model, X, y)

            results = {
                'Model Level': level,
                'Predictors': model.model.exog_names[1:],
                'N (observations)': model.nobs,
                'DF (residuals)': model.df_resid,
                'DF (model)': model.df_model,
                'R-squared': model.rsquared,
                'F-value': model.fvalue,
                'P-value (F)': model.f_pvalue,
                'SSR': model.ssr,
                'SSTO': model.centered_tss,
                'MSE (model)': model.mse_model,
                'MSE (residuals)': model.mse_resid,
                'MSE (total)': model.mse_total,
                'Beta coefs': model.params.to_dict(),
                'P-values (beta coefs)': model.pvalues.to_dict()
            }
            results.update(add_stats)
            if level in change_stats:
                results.update(change_stats[level])
            report.append(results)

        return pd.DataFrame(report)
    
    def diagnostics(self, verbose=True):
        """Performs diagnostics tests on the fitted models.

        Returns:
            dict: A dictionary containing diagnostics results.
        """
        model_results = self.fit_models()
        diagnostics_results = {}

        for level, model in model_results.items():
            X = self.data[model.model.exog_names[1:]]  # Exclude 'const'
            y = self.data[[self.outcome_var]]
            num_ivs = len(self.models[level])

            # Independence of residuals (Durbin-Watson)
            dw_stat, dw_pass = diagnostic_tests.test_durbin_watson(model.resid)
            # Linearity (Pearson's r and Rainbow test)
            pearsons_r_results = diagnostic_tests.test_pearsons_r(X, y)
            rainbow_stat, rainbow_pal, rainbow_pass = diagnostic_tests.test_rainbow(model)
            # Homoscedasticity (Breusch-Pagan and Goldfeld-Quandt tests)
            bp_stat, bp_pval, bp_pass = diagnostic_tests.test_breusch_pagan(model)
            gq_stat, gq_pval, gq_pass = diagnostic_tests.test_goldfeld_quandt(model)
            # Multicollinearity (Pairwise correlations and VIF)
            pairwise_corr, multicollinearity_pass = diagnostic_tests.test_pairwise_correlations(X)
            vif_dict, vif_pass = diagnostic_tests.test_vif(X)
            # Outliers (Extreme Standardized Residuals and High Cook's Distance)
            extreme_resid_indices, extreme_resid_pass = diagnostic_tests.test_extreme_standardized_residuals(model)
            high_cooks_indices, cooks_pass = diagnostic_tests.test_cooks_distance(model, num_ivs)
            # Normality (Mean of Residuals and Shapiro-Wilk Test)
            mean_residuals, mean_residuals_pass = diagnostic_tests.test_mean_of_residuals(model.resid)
            shapiro_stat, shapiro_pval, shapiro_pass = diagnostic_tests.test_shapiro_wilk(model.resid)

            diagnostics_results[level] = {
                'Independence of residuals (Durbin-Watson test)': {'DW stat': dw_stat, 'Passed': dw_pass},
                'Linearity (Pearson r)': pearsons_r_results,
                'Linearity (Rainbow test)': {'Rainbow Stat': rainbow_stat, 'p-value': rainbow_pal, 'Passed': rainbow_pass},
                'Homoscedasticity (Breusch-Pagan test)': {'Lagrange Stat': bp_stat, 'p-value': bp_pval, 'Passed': bp_pass},
                'Homoscedasticity (Goldfeld-Quandt test)': {'F-Stat': gq_stat, 'p-value': gq_pval, 'Passed': gq_pass},
                'Multicollinearity (pairwise correlations)': {'Correlations': pairwise_corr, 'Passed': multicollinearity_pass},
                'Multicollinearity (Variance Inflation Factors)': {'VIFs': vif_dict, 'Passed': vif_pass},
                'Outliers (extreme standardized residuals)': {'Indices': extreme_resid_indices, 'Passed': extreme_resid_pass},
                'Outliers (high Cooks distance)': {'Indices': high_cooks_indices, 'Passed': cooks_pass},
                'Normality (mean of residuals)': {'Mean': mean_residuals, 'Passed': mean_residuals_pass},
                'Normality (Shapiro-Wilk test)': {'SW Stat': shapiro_stat, 'p-value': shapiro_pval, 'Passed': shapiro_pass}
            }
        
        if verbose:
            for level, results in diagnostics_results.items():
                print(f"\nModel Level {level} Diagnostics:")
                for test_name, test_results in results.items():
                    print(f"  {test_name}:")
                    for key, value in test_results.items():
                        print(f"    {key}: {value}")

        return diagnostics_results
    
    def plot_studentized_residuals_vs_fitted(self):
        """Plots studentized residuals against fitted values for all model levels."""
        model_results = self.fit_models()
        num_levels = len(model_results)
        
        fig, axs = plt.subplots(num_levels, 1, figsize=(8, 4 * num_levels))
        
        if num_levels == 1:
            axs = [axs]  # Ensure axs is iterable even for a single subplot

        for ax, (level, model) in zip(axs, model_results.items()):
            plots.create_subplot_residuals_vs_fitted(ax, model, level)

        plt.tight_layout()
        plt.show()
    
    def plot_qq_residuals(self):
        """Plots Normal QQ Plots for all model levels."""
        model_results = self.fit_models()
        fig, axs = plt.subplots(len(model_results), 1, figsize=(8, 4 * len(model_results)))
        if len(model_results) == 1:
            axs = [axs]

        for ax, (level, model) in zip(axs, model_results.items()):
            plots.create_subplot_qq_residuals(model.resid, ax, level)

        plt.tight_layout()
        plt.show()

    def plot_influence(self):
        """Plots Influence Plots for all model levels."""
        model_results = self.fit_models()
        fig, axs = plt.subplots(len(model_results), 1, figsize=(8, 4 * len(model_results)))
        if len(model_results) == 1:
            axs = [axs]

        for ax, (level, model) in zip(axs, model_results.items()):
            plots.create_subplot_influence(model, ax, level)

        plt.tight_layout()
        plt.show()

    def plot_std_residuals(self):
        """Plots Box Plots of Standardized Residuals for all model levels."""
        model_results = self.fit_models()
        fig, axs = plt.subplots(len(model_results), 1, figsize=(8, 4 * len(model_results)))
        if len(model_results) == 1:
            axs = [axs]

        for ax, (level, model) in zip(axs, model_results.items()):
            influence = model.get_influence()
            standardized_residuals = influence.resid_studentized_internal
            plots.create_subplot_std_residuals(standardized_residuals, ax, level)

        plt.tight_layout()
        plt.show()
    
    def plot_histogram_std_residuals(self):
        """Plots Histogram of Standardized Residuals for all model levels."""
        model_results = self.fit_models()
        fig, axs = plt.subplots(len(model_results), 1, figsize=(8, 4 * len(model_results)))
        if len(model_results) == 1:
            axs = [axs]

        for ax, (level, model) in zip(axs, model_results.items()):
            influence = model.get_influence()
            standardized_residuals = influence.resid_studentized_internal
            plots.create_subplot_histogram_std_residuals(standardized_residuals, ax, level)

        plt.tight_layout()
        plt.show()

    def plot_partial_regression(self):
        """Plots Partial Regression Plots for all model levels."""
        model_results = self.fit_models()
        num_ivs = max(len(ivs) for ivs in self.models.values())
        fig_size = (15, min(10, 5 * num_ivs))

        for level, model in model_results.items():
            plots.create_subplot_partial_regression(model, fig_size, level)


import pandas as pd
import matplotlib.pyplot as plt
import seaborn as sns
import statsmodels.api as sm

def create_subplot_residuals_vs_fitted(ax, model, level):
    influence = model.get_influence()
    student_resid = influence.resid_studentized_internal
    fitted_vals = model.fittedvalues
    df_residfitted = pd.DataFrame({'Studentized Residuals': student_resid, 'Fitted Values': fitted_vals})

    sns.residplot(ax=ax, data=df_residfitted, x='Fitted Values', y='Studentized Residuals', lowess=True,
                  scatter_kws={'alpha': 0.8}, line_kws={'color': 'red', 'lw': 1, 'alpha': 1})
    ax.set(ylim=(-3.5, 3.5))
    ax.set_title(f'Residuals vs Fitted (Model Level {level})')
    ax.set_xlabel('Fitted values')
    ax.set_ylabel('Studentized Residuals')

def create_subplot_qq_residuals(residuals, ax, level):
    sm.qqplot(residuals, line='45', fit=True, ax=ax)
    ax.set_title(f'Normal QQ Plot of Residuals (Model Level {level})')

def create_subplot_influence(model, ax, level):
    sm.graphics.influence_plot(model, ax=ax, criterion="cooks")
    ax.set_title(f'Influence Plot (Model Level {level})')
    ax.set_xlabel('H Leverage')
    ax.set_ylabel('Studentized Residuals')

def create_subplot_std_residuals(residuals, ax, level):
    sns.boxplot(y=residuals, ax=ax)
    sns.swarmplot(y=residuals, color="red", ax=ax)
    ax.set_title(f'Boxplot of Standardized Residuals (Model Level {level})')
    ax.set_ylabel('Standardized Residuals')
    ax.set_ylim(-3.5, 3.5)

def create_subplot_histogram_std_residuals(residuals, ax, level):
    sns.histplot(residuals, kde=True, bins=16, ax=ax)
    ax.set_title(f'Histogram of Standardized Residuals (Model Level {level})')
    ax.set_xlabel('Standardized Residuals')
    ax.set_ylabel('Frequency')
    ax.set_xlim(-3.5, 3.5)

def create_subplot_partial_regression(model, fig_size, level):
    fig = plt.figure(figsize=fig_size)
    sm.graphics.plot_partregress_grid(model, fig=fig)
    fig.suptitle(f'Partial Regression Plots (Model Level {level})', y=1)

    plt.tight_layout()
    plt.show()


#!/usr/bin/env python

"""The setup script."""

from setuptools import setup, find_packages

with open('README.md') as readme_file:
    readme = readme_file.read()

with open('HISTORY.rst') as history_file:
    history = history_file.read()

requirements = ['scipy>=1.7.3',
                'pandas>=1.3.5',
                'matplotlib>=3.5.2',
                'seaborn>=0.11',
                'statsmodels>=0.13',
                'pingouin>=0.5.3']
test_requirements = ['pytest>=3']

setup(
    author="Toomas Erik Anijärv",
    author_email='toomaserikanijarv@gmail.com',
    python_requires='>=3.9',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
    ],
    description="HLR - Hierarchical Linear Regression for Python",
    install_requires=requirements,
    license="GNU General Public License v3",
    long_description_content_type='text/markdown',
    long_description=readme + '\n\n' + history,
    include_package_data=True,
    keywords='HLR',
    name='HLR',
    packages=find_packages(include=['HLR', 'HLR.*']),
    test_suite='tests',
    tests_require=test_requirements,
    url='https://github.com/teanijarv/HLR',
    version='0.2.0',
    zip_safe=False,
)



"""Functions for diagnostic tests to test for assumptions.

Authors
-------
Toomas Erik Anijärv toomaserikanijarv@gmail.com github.com/teanijarv
Rory Boyle rorytboyle@gmail.com github.com/rorytboyle
"""

import scipy.stats
import numpy as np
from statsmodels.stats.stattools import durbin_watson
from statsmodels.stats import diagnostic as sm_diagnostic
from statsmodels.stats.outliers_influence import variance_inflation_factor
from statsmodels.tools.tools import add_constant

def test_durbin_watson(residuals):
    """Tests the independence of residuals using the Durbin-Watson statistic.

    Args:
        residuals (pd.Series): Residuals from a regression model.

    Returns:
        tuple: A tuple containing the Durbin-Watson statistic and a boolean indicating if the test is passed.
    """
    dw_stat = durbin_watson(residuals, axis=0)
    dw_pass = 1.5 < dw_stat < 2.5
    return dw_stat, dw_pass

def test_pearsons_r(X, y):
    """Tests linearity between each independent variable and the dependent variable using Pearson's r.

    Args:
        X (pd.DataFrame): Independent variables.
        y (pd.Series): Dependent variable.

    Returns:
        dict: A dictionary containing Pearson's r, p-value, and pass status for each independent variable.
    """
    linearity_results = {}
    for var in X.columns:
        r_value, p_value = scipy.stats.pearsonr(X[var], y.iloc[:, 0])
        test_passed = p_value < 0.05
        linearity_results[var] = {'Pearson r': r_value, 'p-value': p_value, 'Passed': test_passed}
    return linearity_results

def test_rainbow(model):
    """Tests linearity between dependent variable and independent variables collectively using Rainbow test.

    Args:
        model (RegressionResults): The fitted OLS model.

    Returns:
        tuple: A tuple containing the Rainbow statistic, p-value, and pass status.
    """
    rainbow_stat, p_value = sm_diagnostic.linear_rainbow(model)
    test_passed = p_value > 0.05
    return rainbow_stat, p_value, test_passed

def test_breusch_pagan(model):
    """Performs the Breusch-Pagan test for homoscedasticity.

    Args:
        model (RegressionResults): The fitted OLS model.

    Returns:
        tuple: A tuple containing lagrange multiplier statistic, p-value and pass status for the Breusch-Pagan test.
    """
    bp_test = sm_diagnostic.het_breuschpagan(model.resid, model.model.exog)
    bp_test_passed = bp_test[1] > 0.05
    return bp_test[0], bp_test[1], bp_test_passed

def test_goldfeld_quandt(model):
    """Performs the Goldfeld-Quandt test for homoscedasticity.

    Args:
        model (RegressionResults): The fitted OLS model.

    Returns:
        tuple: A tuple containing the F-statistic, p-value and pass status for the Goldfeld-Quandt test.
    """
    gq_test = sm_diagnostic.het_goldfeldquandt(model.resid, model.model.exog)
    gq_test_passed = gq_test[1] > 0.05
    return gq_test[0], gq_test[1], gq_test_passed

def test_pairwise_correlations(X):
    """Tests for multicollinearity by checking pairwise correlations between independent variables.

    Args:
        X (pd.DataFrame): Independent variables.

    Returns:
        tuple: A tuple containing a dictionary of pairwise correlations and a boolean indicating if the test is passed.
    """
    if len(X.columns) > 1:
        pairwise_corr = X.corr()
        np.fill_diagonal(pairwise_corr.values, np.nan)  # Excluding self-correlation
        high_corr = pairwise_corr.abs() >= 0.7
        test_passed = not high_corr.values[np.triu_indices_from(high_corr, k=1)].any()

        # Convert correlation matrix to a more readable dictionary format (upper triangle only)
        corr_dict = {}
        for i, col1 in enumerate(pairwise_corr.columns):
            for j, col2 in enumerate(pairwise_corr.columns):
                if i < j and not np.isnan(pairwise_corr.at[col1, col2]):
                    pair = f"{col1}-{col2}"
                    corr_dict[pair] = pairwise_corr.at[col1, col2]

    else:
        # Test automatically passed if there is only one independent variable
        corr_dict = {}
        test_passed = True

    return corr_dict, test_passed

def test_vif(X):
    """Calculates Variance Inflation Factors for each independent variable.

    Args:
        X (pd.DataFrame): Independent variables.

    Returns:
        tuple: A tuple containing a dictionary of VIFs and a boolean indicating if the test is passed.
    """
    if len(X.columns) > 1:
        X_const = add_constant(X)
        vifs = {X_const.columns[i]: variance_inflation_factor(X_const.values, i) for i in range(1, X_const.shape[1])}
        test_passed = all(vif < 5 for vif in vifs.values())
    else:
        vifs = {}
        test_passed = True

    return vifs, test_passed

def test_extreme_standardized_residuals(model):
    """Identifies extreme standardized residuals.

    Args:
        model (RegressionResults): The fitted OLS model.

    Returns:
        list: A list of indices with extreme standardized residuals.
    """
    influence = model.get_influence()
    standardized_residuals = influence.resid_studentized_internal
    extreme_residuals_indices = np.where(np.abs(standardized_residuals) > 3)[0]
    test_passed = len(extreme_residuals_indices) == 0
    return extreme_residuals_indices, test_passed

def test_cooks_distance(model, num_ivs):
    """Identifies high Cook's distance values.

    Args:
        model (RegressionResults): The fitted OLS model.
        num_ivs (int): Number of independent variables in the model.

    Returns:
        list: A list of indices with high Cook's distance.
    """
    influence = model.get_influence()
    cooks_d = influence.cooks_distance[0]

    # Setting Cook's distance threshold based on the number of IVs
    cooks_cutoff = 0.7 if num_ivs == 1 else 0.8 if num_ivs == 2 else 0.85
    high_cooks_indices = np.where(cooks_d > cooks_cutoff)[0]
    test_passed = len(high_cooks_indices) == 0
    return high_cooks_indices, test_passed

def test_mean_of_residuals(residuals):
    """Checks if the mean of the residuals is approximately zero.

    Args:
        residuals (pd.Series): Residuals from a regression model.

    Returns:
        tuple: A tuple containing the mean of residuals and a boolean indicating if the test is passed.
    """
    mean_residuals = residuals.mean()
    test_passed = -0.1 < mean_residuals < 0.1
    return mean_residuals, test_passed

def test_shapiro_wilk(residuals):
    """Performs the Shapiro-Wilk test for normality.

    Args:
        residuals (pd.Series): Residuals from a regression model.

    Returns:
        tuple: A tuple containing the p-value and pass status for the Shapiro-Wilk test.
    """
    shapiro_stat, shapiro_p_value = scipy.stats.shapiro(residuals)
    test_passed = shapiro_p_value > 0.05
    return shapiro_stat, shapiro_p_value, test_passed